#include <stdio.h>
#include "action.h"
#include "defs.h"
#include "dep.h"
#include "files.h"
#include "gram.h"
#include "new.h"
#include "state.h"
#include "tokens.h"

action **parser;
int SRtotal;
int RRtotal;
short *SRconflicts;
short *RRconflicts;
short *defred;

LOCAL int SRcount;
LOCAL int RRcount;
LOCAL int gotosetsize;
LOCAL int tokensetsize;
LOCAL unsigned *vgotos;
LOCAL short *rule_count;

extern action *parse_actions();
extern action *get_shifts();
extern action *add_reductions();
extern action *add_reduce();


make_parser()
{
  register int i;

  parser = NEW2(nstates, action *);
  for (i = 0; i < final_state; i++)
    parser[i] = parse_actions(i);

  remove_conflicts();
  if (SRtotal + RRtotal > 0)
    total_conflicts();

  defreds();
}

action *
parse_actions(stateno)
register int stateno;
{
  register action *actions;

  actions = get_shifts(stateno);
  actions = add_reductions(stateno, actions);
  return (actions);
}

action *
get_shifts(stateno)
int stateno;
{
  register action *actions, *temp;
  register shifts *sp;
  register short *to_state;
  register int i, k;
  register int symbol;

  actions = NIL(action);
  sp = shift_table[stateno];
  if (sp)
    {
      to_state = sp->shifts;
      for (i = sp->nshifts - 1; i >= 0; i--)
	{
	  k = to_state[i];
	  symbol = accessing_symbol[k];
	  if (ISTOKEN(symbol))
	    {
	      temp = NEW(action);
	      temp->next = actions;
	      temp->symbol = symbol;
	      temp->number = k;
	      temp->prec = symbol_prec[symbol];
	      temp->action_code = SHIFT;
	      temp->assoc = symbol_assoc[symbol];
	      actions = temp;
	    }
	}
    }
  return (actions);
}

action *
add_reductions(stateno, actions)
int stateno;
register action *actions;
{
  register int i, j, m, n;
  register int ruleno, tokensetsize;
  register unsigned *rowp;

  tokensetsize = WORDSIZE(ntokens);
  m = lookaheads[stateno];
  n = lookaheads[stateno + 1];
  for (i = m; i < n; i++)
    {
      ruleno = LAruleno[i];
      rowp = LA + i * tokensetsize;
      for (j = ntokens - 1; j >= 0; j--)
	{
	  if (BIT(rowp, j))
	    actions = add_reduce(actions, ruleno, j);
	}
    }
  return (actions);
}

action *
add_reduce(actions, ruleno, symbol)
register action *actions;
register int ruleno, symbol;
{
  register action *temp, *prev, *next;

  prev = NIL(action);
  next = actions;
  while (next && next->symbol < symbol)
    {
      prev = next;
      next = next->next;
    }

  while (next && next->symbol == symbol && next->action_code == SHIFT)
    {
      prev = next;
      next = next->next;
    }

  while (next && next->symbol == symbol &&
 	 next->action_code == REDUCE && next->number < ruleno)
    {
      prev = next;
      next = next->next;
    }

  temp = NEW(action);
  temp->next = next;
  temp->symbol = symbol;
  temp->number = ruleno;
  temp->prec = rprec[ruleno];
  temp->action_code = REDUCE;
  temp->assoc = rassoc[ruleno];

  if (prev)
    prev->next = temp;
  else
    actions = temp;

  return (actions);
}

remove_conflicts()
{
  register int i;
  register int symbol;
  register action *p, *q;

  SRtotal = 0;
  RRtotal = 0;
  SRconflicts = NEW2(nstates, short);
  RRconflicts = NEW2(nstates, short);
  for (i = 0; i < nstates; i++)
    {
      SRcount = 0;
      RRcount = 0;
      p = parser[i];
      while (p)
	{
	  symbol = p->symbol;
	  q = p;
	  while (q->next && q->next->symbol == symbol)
	    q = q->next;
	  if (p != q)
	    resolve_conflicts(i, p, q);
	  p = q->next;
	}
      SRtotal += SRcount;
      RRtotal += RRcount;
      SRconflicts[i] = SRcount;
      RRconflicts[i] = RRcount;
    }
}

resolve_conflicts(stateno, first, last)
int stateno;
register action *first, *last;
{
  register action *p;
  register int count;
  static char *myname[] = "resolve_conflicts";

  count = 1;
  p = first;
  while (p != last)
    {
      count++;
      p = p->next;
    }
  if (count < 2)
    panic(myname);

  if (first->action_code == SHIFT && count == 2 &&
      first->prec > 0 && last->prec > 0)
    {
      if (first->prec == last->prec)
	{
	  if (first->assoc == LEFT)
	    first->suppressed = 2;
	  else if (first->assoc == RIGHT)
	    last->suppressed = 2;
	  else
	    {
	      first->suppressed = 2;
	      last->suppressed = 2;
	      first->action_code = ERROR;
	      last->action_code = ERROR;
	    }
	}
      else if (first->prec < last->prec)
	first->suppressed = 2;
      else
	last->suppressed = 2;
    }
  else
    {
      if (first->action_code == SHIFT)
	SRcount++;
      else
	RRcount++;
      p = first;
      while (p != last)
	{
	  p = p->next;
	  p->suppressed = 1;
	}
    }
}

total_conflicts()
{
  fprintf(stderr, "%s contains ", input_file_name);

  if (SRtotal == 1)
    fprintf(stderr, "1 shift/reduce conflict");
  else
    fprintf(stderr, "%d shift/reduce conflicts", SRtotal);

  if (SRtotal && RRtotal)
    fprintf(stderr, " and ");

  if (RRtotal == 1)
    fprintf(stderr, "1 reduce/reduce conflict");
  else if (RRtotal > 1)
    fprintf(stderr, "%d reduce/reduce conflicts", RRtotal);

  putc('.', stderr);
  putc(NEWLINE, stderr);
}

find_defred(stateno)
int stateno;
{
  register int i, max, defred;
  register action *p;

  static char myname[] = "find_defered";

  for (i = 0; i < nrules; i++)
    rule_count[i] = 0;

  for (p = parser[stateno]; p; p = p->next)
    {
      if (p->action_code == ERROR)
	{
	  if (p == 0)
	    panic(myname);
	  p = p->next;
	  if (p->action_code != ERROR)
	    panic(myname);
	  rule_count[p->number] == -1;
	}
      else if (p->action_code == REDUCE && p->suppressed == 0 &&
	       rule_count[p->number] >= 0)
	rule_count[p->number]++;
    }

  max = 0;
  defred = -1;
  for (i = 0; i < nrules; i++)
    {
      if (rule_count[i] > max)
	{
	  max = rule_count[i];
	  defred = i;
	}
    }

  return (defred);
}

defreds()
{
  register int i;

  rule_count = NEW2(nrules, short);
  defred = NEW2(nstates, short);

  for (i = 0; i < nstates; i++)
    defred[i] = find_defred(i);

  FREE(rule_count);
}

free_action_row(p)
register action *p;
{
  register action *q;

  while (p)
    {
      q = p->next;
      FREE(p);
      p = q;
    }
}

free_parser()
{
  register int i;

  for (i = 0; i < nstates; i++)
    free_action_row(parser[i]);

  FREE(parser);
}
